import json
import os
import asyncio
import aiohttp
import logging
import traceback
import random
from importlib.metadata import version

from proximl.auth import Auth
from proximl.datasets import Datasets
from proximl.models import Models
from proximl.checkpoints import Checkpoints
from proximl.volumes import Volumes
from proximl.jobs import Jobs
from proximl.gpu_types import GpuTypes
from proximl.environments import Environments
from proximl.exceptions import ApiError, ProxiMLException
from proximl.connections import Connections
from proximl.projects import Projects
from proximl.cloudbender import Cloudbender


async def delayed_close(ws):
    await asyncio.sleep(15)
    if not ws.closed:
        await ws.close()


class ProxiML(object):
    def __init__(self, **kwargs):
        self._version = version("proximl")
        CONFIG_DIR = kwargs.get("config_dir") or os.path.expanduser(
            os.environ.get("PROXIML_CONFIG_DIR") or "~/.proximl"
        )
        try:
            with open(f"{CONFIG_DIR}/environment.json", "r") as file:
                env_str = file.read().replace("\n", "")
            env = json.loads(env_str)
        except OSError:
            env = dict()
        try:
            with open(f"{CONFIG_DIR}/config.json", "r") as file:
                config_str = file.read().replace("\n", "")
            config = json.loads(config_str)
        except OSError:
            config = dict()
        self.domain_suffix = (
            kwargs.get("domain_suffix")
            or os.environ.get("PROXIML_DOMAIN_SUFFIX")
            or env.get("domain_suffix")
            or "proximl.ai"
        )
        self.auth = Auth(
            config_dir=CONFIG_DIR,
            domain_suffix=self.domain_suffix,
            user=kwargs.get("user"),
            key=kwargs.get("key"),
            region=kwargs.get("region"),
            client_id=kwargs.get("client_id"),
            pool_id=kwargs.get("pool_id"),
        )
        self.active_project = (
            kwargs.get("project")
            or os.environ.get("PROXIML_PROJECT")
            or config.get("project")
        )
        self.datasets = Datasets(self)
        self.models = Models(self)
        self.checkpoints = Checkpoints(self)
        self.volumes = Volumes(self)
        self.jobs = Jobs(self)
        self.gpu_types = GpuTypes(self)
        self.environments = Environments(self)
        self.connections = Connections(self)
        self.projects = Projects(self)
        self.cloudbender = Cloudbender(self)
        self.api_url = (
            kwargs.get("api_url")
            or os.environ.get("PROXIML_API_URL")
            or env.get("api_url")
            or f"api.{self.domain_suffix}"
        )
        self.ws_url = (
            kwargs.get("ws_url")
            or os.environ.get("PROXIML_WS_URL")
            or env.get("ws_url")
            or f"api-ws.{self.domain_suffix}"
        )

    @property
    def project(self) -> str:
        return self.active_project

    async def _query(self, path, method, params=None, data=None, headers=None,max_retries=3, backoff_factor=0.5):
        try:
            tokens = self.auth.get_tokens()
        except ProxiMLException as e:
            raise e
        except Exception:
            raise ProxiMLException(
                f"Error getting authorization tokens.  Verify configured credentials. Error: {traceback.format_exc()}"
            )
        logging.debug(
            f"Call parameters - Path: {path}, Method: {method}, Params: {params}, Body: {data}, Headers: {headers}"
        )
        headers = (
            {
                **headers,
                **{
                    "Authorization": tokens.get("id_token"),
                    "User-Agent": f"proxiML-sdk/{self._version}",
                },
            }
            if headers
            else {
                "Authorization": tokens.get("id_token"),
                "User-Agent": f"proxiML-sdk/{self._version}",
            }
        )
        if params:
            if not isinstance(params, dict):
                raise ProxiMLException("Query parameters must be a valid dictionary")
            params = {
                k: (str(v).lower() if isinstance(v, bool) else v)
                for k, v in params.items()
            }  ## aiohttp doesn't support boolean
        if (
            method != "POST"
            and self.active_project
            and (not params or "project_uuid" not in params.keys())
        ):
            params = (
                {**params, **{"project_uuid": self.active_project}}
                if params
                else {"project_uuid": self.active_project}
            )

        if "Content-Type" not in headers:
            headers["Content-Type"] = "application/json"
        url = f"https://{self.api_url}{path}"

        logging.debug(
            f"Request - Url: {url}, Method: {method}, Params: {params}, Body: {data}, Headers: {headers}"
        )
        for attempt in range(max_retries):
            try:
                async with aiohttp.ClientSession() as session:
                    async with session.request(
                        method,
                        url,
                        data=json.dumps(data),
                        headers=headers,
                        params=params,
                    ) as resp:
                        if (resp.status // 100) in [4, 5]:
                            if resp.status == 502 and attempt < max_retries - 1:
                                wait_time = (2 ** attempt) * backoff_factor * (random.random() + 0.5)
                                await asyncio.sleep(wait_time)
                                continue
                            else:
                                what = await resp.read()
                                content_type = resp.headers.get("content-type", "")
                                resp.close()
                                if content_type == "application/json":
                                    raise ApiError(resp.status, json.loads(what.decode("utf8")))
                                else:
                                    raise ApiError(resp.status, {"message": what.decode("utf8")})
                        results = await resp.json()
                        return results
            except aiohttp.ClientResponseError as e:
                if e.status == 502 and attempt < max_retries - 1:
                    wait_time = (2 ** attempt) * backoff_factor * (random.random() + 0.5)
                    await asyncio.sleep(wait_time)
                    continue
                else:
                    raise  ApiError(e.status, f"Error {e.message}")

        raise ProxiMLException("Unexpected API failure")

    async def _ws_subscribe(self, entity, project_uuid, id, msg_handler):
        headers = {
            "User-Agent": f"proxiML-sdk/{self._version}",
            "Content-Type": "application/json",
        }
        try:
            tokens = self.auth.get_tokens()
        except ProxiMLException as e:
            raise e
        except Exception:
            raise ProxiMLException(
                f"Error getting authorization tokens.  Verify configured credentials. Error: {traceback.format_exc()}"
            )
        async with aiohttp.ClientSession() as session:
            done = False
            async with session.ws_connect(
                f"wss://{self.ws_url}?Authorization={tokens.get('id_token')}",
                headers=headers,
                heartbeat=30,
            ) as ws:
                asyncio.create_task(
                    ws.send_json(
                        dict(
                            action="getlogs",
                            data=dict(
                                type="init",
                                entity=entity,
                                id=id,
                                project_uuid=project_uuid,
                            ),
                        )
                    )
                )
                asyncio.create_task(
                    ws.send_json(
                        dict(
                            action="subscribe",
                            data=dict(
                                type="logs",
                                entity=entity,
                                id=id,
                                project_uuid=project_uuid,
                            ),
                        )
                    )
                )
                async for msg in ws:
                    if msg.type in (
                        aiohttp.WSMsgType.CLOSED,
                        aiohttp.WSMsgType.ERROR,
                        aiohttp.WSMsgType.CLOSE,
                    ):
                        logging.debug(
                            f"Websocket Received Closed Message.  Done? {done}"
                        )
                        await ws.close()
                        break
                    data = json.loads(msg.data)
                    if data.get("type") == "end":
                        done = True
                        asyncio.create_task(delayed_close(ws))
                    else:
                        msg_handler(data)
            logging.debug(f"Websocket Disconnected.  Done? {done}")

            connection_tries = 0
            while not done:
                tokens = self.auth.get_tokens()
                try:
                    async with session.ws_connect(
                        f"wss://{self.ws_url}?Authorization={tokens.get('id_token')}",
                        headers=headers,
                        heartbeat=30,
                    ) as ws:
                        asyncio.create_task(
                            ws.send_json(
                                dict(
                                    action="subscribe",
                                    data=dict(
                                        type="logs",
                                        entity=entity,
                                        id=id,
                                        project_uuid=project_uuid,
                                    ),
                                )
                            )
                        )
                        async for msg in ws:
                            if msg.type in (
                                aiohttp.WSMsgType.CLOSED,
                                aiohttp.WSMsgType.ERROR,
                                aiohttp.WSMsgType.CLOSE,
                            ):
                                logging.debug(
                                    f"Websocket Received Closed Message.  Done? {done}"
                                )
                                await ws.close()
                                break
                            data = json.loads(msg.data)
                            if data.get("type") == "end":
                                done = True
                                asyncio.create_task(delayed_close(ws))
                            else:
                                msg_handler(data)
                    connection_tries = 0
                    logging.debug(f"Websocket Disconnected.  Done? {done}")
                except Exception as e:
                    connection_tries += 1
                    logging.debug(f"Connection error: {traceback.format_exc()}")
                    if connection_tries == 5:
                        raise ApiError(
                            500,
                            {"message": f"Connection error: {traceback.format_exc()}"},
                        )

    def set_active_project(self, project_uuid):
        CONFIG_DIR = os.path.expanduser(
            os.environ.get("PROXIML_CONFIG_DIR") or "~/.proximl"
        )
        with open(f"{CONFIG_DIR}/config.json", "w") as file:
            json.dump(dict(project=project_uuid), file)



import json
import logging
import asyncio
import math

from proximl.exceptions import (
    ApiError,
    SpecificationError,
    ProxiMLException,
)


class Datastores(object):
    def __init__(self, proximl):
        self.proximl = proximl

    async def get(self, provider_uuid, region_uuid, id, **kwargs):
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/datastore/{id}",
            "GET",
            kwargs,
        )
        return Datastore(self.proximl, **resp)

    async def list(self, provider_uuid, region_uuid, **kwargs):
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/datastore",
            "GET",
            kwargs,
        )
        datastores = [Datastore(self.proximl, **datastore) for datastore in resp]
        return datastores

    async def create(
        self,
        provider_uuid,
        region_uuid,
        name,
        type,
        **kwargs,
    ):
        logging.info(f"Creating Datastore {name}")
        data = dict(
            name=name,
            type=type,
            **kwargs,
        )
        payload = {k: v for k, v in data.items() if v is not None}
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/datastore",
            "POST",
            None,
            payload,
        )
        datastore = Datastore(self.proximl, **resp)
        logging.info(f"Created Datastore {name} with id {datastore.id}")
        return datastore

    async def remove(self, provider_uuid, region_uuid, id, **kwargs):
        await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/datastore/{id}",
            "DELETE",
            kwargs,
        )


class Datastore:
    def __init__(self, proximl, **kwargs):
        self.proximl = proximl
        self._datastore = kwargs
        self._id = self._datastore.get("store_id")
        self._provider_uuid = self._datastore.get("provider_uuid")
        self._region_uuid = self._datastore.get("region_uuid")
        self._type = self._datastore.get("type")
        self._name = self._datastore.get("name")

    @property
    def id(self) -> str:
        return self._id

    @property
    def provider_uuid(self) -> str:
        return self._provider_uuid

    @property
    def region_uuid(self) -> str:
        return self._region_uuid

    @property
    def type(self) -> str:
        return self._type

    @property
    def name(self) -> str:
        return self._name

    def __str__(self):
        return json.dumps({k: v for k, v in self._datastore.items()})

    def __repr__(self):
        return f"Datastore( proximl , **{self._datastore.__repr__()})"

    def __bool__(self):
        return bool(self._id)

    async def remove(self):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/datastore/{self._id}",
            "DELETE",
        )

    async def refresh(self):
        resp = await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/datastore/{self._id}",
            "GET",
        )
        self.__init__(self.proximl, **resp)
        return self



import re
import sys
import asyncio
from pytest import mark, fixture

pytestmark = [mark.sdk, mark.integration, mark.projects]


@mark.create
@mark.asyncio
@mark.xdist_group("project_resources")
class ProjectServicesTests:
    @fixture(scope="class")
    async def project_service(self, project):
        services = await project.services.list()
        yield services[0]

    async def test_list_project_services(self, project):
        services = await project.services.list()
        assert len(services) > 0

    async def test_project_service_properties(self, project, project_service):
        assert isinstance(project_service.project_uuid, str)
        assert isinstance(project_service.name, str)
        assert isinstance(project_service.id, str)
        assert isinstance(project_service.hostname, str)
        assert isinstance(project_service.region_uuid, str)
        assert isinstance(project_service.public, bool)
        assert project.id == project_service.project_uuid

    async def test_project_service_str(self, project_service):
        string = str(project_service)
        regex = r"^{.*\"name\": \"" + project_service.name + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    async def test_project_service_repr(self, project_service):
        string = repr(project_service)
        regex = (
            r"^ProjectService\( proximl , \*\*{.*'name': '"
            + project_service.name
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)



import json
import logging


class ProjectServices(object):
    def __init__(self, proximl, project_id):
        self.proximl = proximl
        self.project_id = project_id

    async def get(self, id, **kwargs):
        resp = await self.proximl._query(
            f"/project/{self.project_id}/services/{id}", "GET", kwargs
        )
        return ProjectService(self.proximl, **resp)

    async def list(self, **kwargs):
        resp = await self.proximl._query(
            f"/project/{self.project_id}/services", "GET", kwargs
        )
        services = [ProjectService(self.proximl, **service) for service in resp]
        return services

    async def refresh(self):
        await self.proximl._query(f"/project/{self.project_id}/services", "PATCH")


class ProjectService:
    def __init__(self, proximl, **kwargs):
        self.proximl = proximl
        self._entity = kwargs
        self._id = self._entity.get("id")
        self._project_uuid = self._entity.get("project_uuid")
        self._name = self._entity.get("name")
        self._hostname = self._entity.get("hostname")
        self._public = self._entity.get("public")
        self._region_uuid = self._entity.get("region_uuid")

    @property
    def id(self) -> str:
        return self._id

    @property
    def project_uuid(self) -> str:
        return self._project_uuid

    @property
    def name(self) -> str:
        return self._name

    @property
    def hostname(self) -> str:
        return self._hostname

    @property
    def public(self) -> bool:
        return self._public

    @property
    def region_uuid(self) -> str:
        return self._region_uuid

    def __str__(self):
        return json.dumps({k: v for k, v in self._entity.items()})

    def __repr__(self):
        return f"ProjectService( proximl , **{self._entity.__repr__()})"

    def __bool__(self):
        return bool(self._id)

    async def enable(self):
        await self.proximl._query(
            f"/project/{self._project_uuid}/services/{self._id}/enable", "PATCH"
        )

    async def disable(self):
        await self.proximl._query(
            f"/project/{self._project_uuid}/services/{self._id}/disable", "PATCH"
        )

    async def get_service_ca_certificate(self, **kwargs):
        certificate = await self.proximl._query(
            f"/project/{self._project_uuid}/services/{self._id}/certificate/ca",
            "GET",
            kwargs,
        )
        return certificate
    
    async def sign_client_certificate(self, csr, **kwargs):
        certificate = await self.proximl._query(
            f"/project/{self._project_uuid}/services/{self._id}/certificate/sign",
            "POST",
            kwargs,
            dict(csr=csr)
        )
        return certificate



import json
import logging


class DeviceConfigs(object):
    def __init__(self, proximl):
        self.proximl = proximl

    async def get(self, provider_uuid, region_uuid, id, **kwargs):
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/device/config/{id}",
            "GET",
            kwargs,
        )
        return DeviceConfig(self.proximl, **resp)

    async def list(self, provider_uuid, region_uuid, **kwargs):
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/device/config",
            "GET",
            kwargs,
        )
        device_configs = [
            DeviceConfig(self.proximl, **device_config)
            for device_config in resp
        ]
        return device_configs

    async def create(
        self,
        provider_uuid,
        region_uuid,
        name,
        model_uuid,
        image,
        command,
        **kwargs,
    ):
        logging.info(f"Creating Device Config {name}")
        data = dict(
            name=name,
            model_uuid=model_uuid,
            image=image,
            command=command,
            **kwargs,
        )
        if not data.get("model_project_uuid"):
            data["model_project_uuid"] = self.proximl.active_project

        payload = {k: v for k, v in data.items() if v is not None}
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/device/config",
            "POST",
            None,
            payload,
        )
        device_config = DeviceConfig(self.proximl, **resp)
        logging.info(
            f"Created Device Config {name} with id {device_config.id}"
        )
        return device_config

    async def remove(self, provider_uuid, region_uuid, id, **kwargs):
        await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/device/config/{id}",
            "DELETE",
            kwargs,
        )


class DeviceConfig:
    def __init__(self, proximl, **kwargs):
        self.proximl = proximl
        self._device_config = kwargs
        self._id = self._device_config.get("config_id")
        self._provider_uuid = self._device_config.get("provider_uuid")
        self._region_uuid = self._device_config.get("region_uuid")
        self._name = self._device_config.get("name")

    @property
    def id(self) -> str:
        return self._id

    @property
    def provider_uuid(self) -> str:
        return self._provider_uuid

    @property
    def region_uuid(self) -> str:
        return self._region_uuid

    @property
    def name(self) -> str:
        return self._name

    def __str__(self):
        return json.dumps({k: v for k, v in self._device_config.items()})

    def __repr__(self):
        return f"DeviceConfig( proximl , **{self._device_config.__repr__()})"

    def __bool__(self):
        return bool(self._id)

    async def remove(self):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/device/config/{self._id}",
            "DELETE",
        )

    async def refresh(self):
        resp = await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/device/config/{self._id}",
            "GET",
        )
        self.__init__(self.proximl, **resp)
        return self




from pytest import fixture, mark


pytestmark = [mark.integration,mark.cloudbender]

@fixture(scope="session")
@mark.create
@mark.asyncio
@mark.xdist_group("cloudbender_resources")
async def provider( proximl):
    provider = await proximl.cloudbender.providers.enable(type="test")
    await provider.wait_for("ready")
    yield provider
    await provider.remove()

@fixture(scope="session")
@mark.create
@mark.asyncio
@mark.xdist_group("cloudbender_resources")
async def region(proximl, provider):
    region = await proximl.cloudbender.regions.create(provider_uuid=provider.id,name="test-region",
        public=False,
        storage=dict(mode="local"),)
    await region.wait_for("healthy")
    yield region
    await region.remove()
    await region.wait_for("archived")


import re
import json
import logging
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises
from aiohttp import WSMessage, WSMsgType

import proximl.cloudbender.devices as specimen
from proximl.exceptions import (
    ApiError,
    SpecificationError,
    ProxiMLException,
)

pytestmark = [mark.sdk, mark.unit, mark.cloudbender, mark.devices]


@fixture
def devices(mock_proximl):
    yield specimen.Devices(mock_proximl)


@fixture
def device(mock_proximl):
    yield specimen.Device(
        mock_proximl,
        provider_uuid="1",
        region_uuid="a",
        device_id="x",
        type="device",
        service="compute",
        friendly_name="hq-orin-01",
        hostname="hq-orin-01",
        status="active",
        online=True,
        maintenance_mode=False,
        job_status="stopped",
        job_last_deployed="2023-06-02T21:22:40.084Z",
        job_config_id="job-id-1",
        job_config_revision="1685740490096",
        device_config_id="conf-id-2",
    )


class RegionsTests:
    @mark.asyncio
    async def test_get_device(
        self,
        devices,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await devices.get("1234", "5687", "91011")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/5687/device/91011", "GET", {}
        )

    @mark.asyncio
    async def test_list_devices(
        self,
        devices,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await devices.list("1234", "5687")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/5687/device", "GET", {}
        )

    @mark.asyncio
    async def test_remove_device(
        self,
        devices,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await devices.remove("1234", "4567", "8910")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/4567/device/8910", "DELETE", {}
        )

    @mark.asyncio
    async def test_create_device(self, devices, mock_proximl):
        requested_config = dict(
            provider_uuid="provider-id-1",
            region_uuid="region-id-1",
            friendly_name="phys-device",
            hostname="phys-device",
            minion_id="asdf",
        )
        expected_payload = dict(
            friendly_name="phys-device",
            hostname="phys-device",
            minion_id="asdf",
            type="device",
            service="compute",
        )
        api_response = {
            "provider_uuid": "provider-id-1",
            "region_uuid": "region-id-1",
            "device_id": "rig-id-1",
            "name": "phys-device",
            "type": "device",
            "service": "compute",
            "status": "new",
            "online": False,
            "maintenance_mode": True,
            "job_status": "stopped",
            "job_last_deployed": "2023-06-02T21:22:40.084Z",
            "job_config_id": "job-id-1",
            "job_config_revision": "1685740490096",
            "device_config_id": "conf-id-1",
            "createdAt": "2020-12-31T23:59:59.000Z",
        }

        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await devices.create(**requested_config)
        mock_proximl._query.assert_called_once_with(
            "/provider/provider-id-1/region/region-id-1/device",
            "POST",
            None,
            expected_payload,
        )
        assert response.id == "rig-id-1"


class deviceTests:
    def test_device_properties(self, device):
        assert isinstance(device.id, str)
        assert isinstance(device.provider_uuid, str)
        assert isinstance(device.region_uuid, str)
        assert isinstance(device.name, str)
        assert isinstance(device.hostname, str)
        assert isinstance(device.status, str)
        assert isinstance(device.online, bool)
        assert isinstance(device.maintenance_mode, bool)
        assert isinstance(device.device_config_id, str)
        assert isinstance(device.job_status, str)
        assert isinstance(device.job_last_deployed, str)
        assert isinstance(device.job_config_id, str)
        assert isinstance(device.job_config_revision, str)

    def test_device_str(self, device):
        string = str(device)
        regex = r"^{.*\"device_id\": \"" + device.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_device_repr(self, device):
        string = repr(device)
        regex = (
            r"^Device\( proximl , \*\*{.*'device_id': '"
            + device.id
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_device_bool(self, device, mock_proximl):
        empty_device = specimen.Device(mock_proximl)
        assert bool(device)
        assert not bool(empty_device)

    @mark.asyncio
    async def test_device_remove(self, device, mock_proximl):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await device.remove()
        mock_proximl._query.assert_called_once_with(
            "/provider/1/region/a/device/x", "DELETE"
        )

    @mark.asyncio
    async def test_device_refresh(self, device, mock_proximl):
        api_response = {
            "provider_uuid": "provider-id-1",
            "region_uuid": "region-id-1",
            "device_id": "device-id-1",
            "name": "phys-device",
            "type": "device",
            "service": "compute",
            "status": "new",
            "online": False,
            "maintenance_mode": True,
            "job_status": "stopped",
            "job_last_deployed": "2023-06-02T21:22:40.084Z",
            "job_config_id": "job-id-1",
            "job_config_revision": "1685740490096",
            "device_config_id": "conf-id-1",
            "createdAt": "2020-12-31T23:59:59.000Z",
        }
        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await device.refresh()
        mock_proximl._query.assert_called_once_with(
            f"/provider/1/region/a/device/x", "GET"
        )
        assert device.id == "device-id-1"
        assert response.id == "device-id-1"

    @mark.asyncio
    async def test_device_toggle_maintenance(self, device, mock_proximl):
        api_response = None
        mock_proximl._query = AsyncMock(return_value=api_response)
        await device.toggle_maintenance()
        mock_proximl._query.assert_called_once_with(
            "/provider/1/region/a/device/x/maintenance", "PATCH"
        )

    @mark.asyncio
    async def test_device_run_action(self, device, mock_proximl):
        api_response = None
        mock_proximl._query = AsyncMock(return_value=api_response)
        await device.run_action(command="report")
        mock_proximl._query.assert_called_once_with(
            "/provider/1/region/a/device/x/action",
            "POST",
            None,
            dict(command="report"),
        )

    @mark.asyncio
    async def test_device_set_config(self, device, mock_proximl):
        api_response = {
            "provider_uuid": "provider-id-1",
            "region_uuid": "region-id-1",
            "device_id": "device-id-1",
            "name": "phys-device",
            "type": "device",
            "service": "compute",
            "status": "new",
            "online": False,
            "maintenance_mode": True,
            "job_status": "stopped",
            "job_last_deployed": "2023-06-02T21:22:40.084Z",
            "job_config_id": "job-id-1",
            "job_config_revision": "1685740490096",
            "device_config_id": "config-id-1",
            "createdAt": "2020-12-31T23:59:59.000Z",
        }
        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await device.set_config(device_config_id="config-id-1")
        mock_proximl._query.assert_called_once_with(
            "/provider/1/region/a/device/x",
            "PATCH",
            None,
            dict(device_config_id="config-id-1"),
        )
        assert device.id == "device-id-1"
        assert response.id == "device-id-1"

    @mark.asyncio
    async def test_device_deploy_endpoint(self, device, mock_proximl):
        api_response = None
        mock_proximl._query = AsyncMock(return_value=api_response)
        await device.deploy_endpoint()
        mock_proximl._query.assert_called_once_with(
            "/provider/1/region/a/device/x/deploy", "PUT"
        )

    @mark.asyncio
    async def test_device_stop_endpoint(self, device, mock_proximl):
        api_response = None
        mock_proximl._query = AsyncMock(return_value=api_response)
        await device.stop_endpoint()
        mock_proximl._query.assert_called_once_with(
            "/provider/1/region/a/device/x/stop", "PUT"
        )



import re
import sys
import asyncio
from pytest import mark, fixture

pytestmark = [mark.sdk, mark.integration, mark.cloudbender, mark.providers]


@mark.create
@mark.asyncio
@mark.xdist_group("cloudbender_resources")
class GetProviderTests:
    async def test_get_providers(self, proximl, provider):
        providers = await proximl.cloudbender.providers.list()
        assert len(providers) > 0

    async def test_get_provider(self, proximl, provider):
        response = await proximl.cloudbender.providers.get(provider.id)
        assert response.id == provider.id

    async def test_provider_properties(self, provider):
        assert isinstance(provider.id, str)
        assert isinstance(provider.type, str)
        assert provider.type == "test"
        assert provider.credits == 0

    async def test_provider_str(self, provider):
        string = str(provider)
        regex = r"^{.*\"provider_uuid\": \"" + provider.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    async def test_provider_repr(self, provider):
        string = repr(provider)
        regex = (
            r"^Provider\( proximl , \*\*{.*'provider_uuid': '"
            + provider.id
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)



import warnings
import logging

with warnings.catch_warnings():
    # this will suppress all warnings in this block
    warnings.filterwarnings("ignore", message="int_from_bytes is deprecated")
    from .proximl import ProxiML

logging.basicConfig(
    format="%(asctime)s.%(msecs)03dZ  %(levelname)s  %(message)s",
    datefmt="%Y-%m-%dT%H:%M:%S",
)
logger = logging.getLogger(__name__)


__version__ = "0.5.17"
__all__ = "ProxiML"



import re
import sys
import asyncio
from pytest import mark, fixture

pytestmark = [mark.sdk, mark.integration, mark.projects]


@mark.create
@mark.asyncio
@mark.xdist_group("project_resources")
class ProjectMembersTests:
    @fixture(scope="class")
    async def project_member(self, project):
        member = await project.members.add("test.account@proximl.ai","read","read","read","read","read")
        yield member
        await project.members.remove("test.account@proximl.ai")

    async def test_list_project_members(self, project):
        members = await project.members.list()
        assert len(members) > 0

    async def test_project_member_properties(self, project, project_member):
        assert isinstance(project_member.id, str)
        assert isinstance(project_member.email, str)
        assert isinstance(project_member.project_uuid, str)
        assert isinstance(project_member.owner, bool)
        assert isinstance(project_member.job, str)
        assert isinstance(project_member.dataset, str)
        assert isinstance(project_member.model, str)
        assert isinstance(project_member.checkpoint, str)
        assert isinstance(project_member.volume, str)
        assert project.id == project_member.project_uuid
        assert project_member.id == "test.account@proximl.ai"

    async def test_project_member_str(self, project_member):
        string = str(project_member)
        regex = r"^{.*\"email\": \"" + project_member.email + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    async def test_project_member_repr(self, project_member):
        string = repr(project_member)
        regex = (
            r"^ProjectMember\( proximl , \*\*{.*'email': '"
            + project_member.email
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)



import re
import sys
import asyncio
from pytest import mark, fixture
from cryptography import x509
from cryptography.x509.oid import NameOID, ExtendedKeyUsageOID
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization, hashes

pytestmark = [mark.sdk, mark.integration, mark.cloudbender, mark.regions]

def get_csr(service_id):
    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=4096
    )
    csr = x509.CertificateSigningRequestBuilder().subject_name(x509.Name([
        x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, "proxiML"),
        x509.NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, service_id),
        x509.NameAttribute(NameOID.COMMON_NAME, "test-client"),  # Client identity
    ])).add_extension(
        x509.ExtendedKeyUsage([
            ExtendedKeyUsageOID.CLIENT_AUTH  # Client authentication usage
        ]),
        critical=True
    ).sign(private_key, hashes.SHA256())
    return csr.public_bytes(serialization.Encoding.PEM).decode("utf-8")

@mark.create
@mark.asyncio
@mark.xdist_group("cloudbender_resources")
class GetServiceTests:
    @fixture(scope="class")
    async def service(self, proximl, region):
        service = await proximl.cloudbender.services.create(
            provider_uuid=region.provider_uuid,
            region_uuid=region.id,
            name="CLI Automated Service",
            type="tcp",
            port="8989",
            public=False,
        )
        await service.wait_for("active")
        yield service
        await service.remove()
        await service.wait_for("archived")

    async def test_get_services(self, proximl, region,service):
        services = await proximl.cloudbender.services.list(provider_uuid=region.provider_uuid, region_uuid=region.id)
        assert len(services) > 0

    async def test_get_service(self, proximl, provider, region, service):
        response = await proximl.cloudbender.services.get(provider.id, region.id, service.id)
        assert response.id == service.id

    async def test_service_properties(self, region, service):
        assert isinstance(service.id, str)
        assert isinstance(service.provider_uuid, str)
        assert isinstance(service.region_uuid, str)
        assert isinstance(service.public, bool)
        assert service.port == "8989"
        assert service.provider_uuid == region.provider_uuid
        assert service.region_uuid == region.id

    async def test_service_str(self, service):
        string = str(service)
        regex = r"^{.*\"service_id\": \"" + service.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    async def test_service_repr(self, service):
        string = repr(service)
        regex = (
            r"^Service\( proximl , \*\*{.*'service_id': '"
            + service.id
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)

    async def test_service_certificate(self, service):
        service = await service.generate_certificate()
        assert isinstance(service._service.get("auth_cert"), str)
        csr = get_csr(service.id)
        certificate = await service.sign_client_certificate(csr)
        assert isinstance(certificate, str)



import json
import logging
import asyncio
import math

from proximl.exceptions import ApiError, SpecificationError, ProxiMLException, NodeError


class Nodes(object):
    def __init__(self, proximl):
        self.proximl = proximl

    async def get(self, provider_uuid, region_uuid, id, **kwargs):
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/node/{id}",
            "GET",
            kwargs,
        )
        return Node(self.proximl, **resp)

    async def list(self, provider_uuid, region_uuid, **kwargs):
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/node",
            "GET",
            kwargs,
        )
        nodes = [Node(self.proximl, **node) for node in resp]
        return nodes

    async def create(
        self,
        provider_uuid,
        region_uuid,
        friendly_name,
        hostname,
        minion_id=None,
        type="permanent",
        service="compute",
        **kwargs,
    ):
        logging.info(f"Creating Node {friendly_name}")
        data = dict(
            friendly_name=friendly_name,
            hostname=hostname,
            minion_id=minion_id,
            type=type,
            service=service,
            **kwargs,
        )
        payload = {k: v for k, v in data.items() if v is not None}
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/node",
            "POST",
            None,
            payload,
        )
        node = Node(self.proximl, **resp)
        logging.info(f"Created Node {friendly_name} with id {node.id}")
        return node

    async def remove(self, provider_uuid, region_uuid, id, **kwargs):
        await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/node/{id}",
            "DELETE",
            kwargs,
        )


class Node:
    def __init__(self, proximl, **kwargs):
        self.proximl = proximl
        self._node = kwargs
        self._id = self._node.get("rig_uuid")
        self._provider_uuid = self._node.get("provider_uuid")
        self._region_uuid = self._node.get("region_uuid")
        self._type = self._node.get("type")
        self._service = self._node.get("service")
        self._name = self._node.get("friendly_name")
        self._hostname = self._node.get("hostname")
        self._status = self._node.get("status")
        self._online = self._node.get("online")
        self._maintenance_mode = self._node.get("maintenance_mode")

    @property
    def id(self) -> str:
        return self._id

    @property
    def provider_uuid(self) -> str:
        return self._provider_uuid

    @property
    def region_uuid(self) -> str:
        return self._region_uuid

    @property
    def type(self) -> str:
        return self._type

    @property
    def service(self) -> str:
        return self._service

    @property
    def name(self) -> str:
        return self._name

    @property
    def hostname(self) -> str:
        return self._hostname

    @property
    def status(self) -> str:
        return self._status

    @property
    def online(self) -> bool:
        return self._online

    @property
    def maintenance_mode(self) -> bool:
        return self._maintenance_mode

    def __str__(self):
        return json.dumps({k: v for k, v in self._node.items()})

    def __repr__(self):
        return f"Node( proximl , **{self._node.__repr__()})"

    def __bool__(self):
        return bool(self._id)

    async def remove(self):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/node/{self._id}",
            "DELETE",
        )

    async def refresh(self):
        resp = await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/node/{self._id}",
            "GET",
        )
        self.__init__(self.proximl, **resp)
        return self

    async def toggle_maintenance(self):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/node/{self._id}/maintenance",
            "PATCH",
        )

    async def run_action(self, command):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/node/{self._id}/action",
            "POST",
            None,
            dict(command=command),
        )

    async def wait_for(self, status, timeout=300):
        if self.status == status:
            return
        valid_statuses = ["active", "maintenance", "offline", "stopped", "archived"]
        if not status in valid_statuses:
            raise SpecificationError(
                "status",
                f"Invalid wait_for status {status}.  Valid statuses are: {valid_statuses}",
            )
        MAX_TIMEOUT = 24 * 60 * 60
        if timeout > MAX_TIMEOUT:
            raise SpecificationError(
                "timeout",
                f"timeout must be less than {MAX_TIMEOUT} seconds.",
            )

        POLL_INTERVAL_MIN = 5
        POLL_INTERVAL_MAX = 60
        POLL_INTERVAL = max(min(timeout / 60, POLL_INTERVAL_MAX), POLL_INTERVAL_MIN)
        retry_count = math.ceil(timeout / POLL_INTERVAL)
        count = 0
        while count < retry_count:
            await asyncio.sleep(POLL_INTERVAL)
            try:
                await self.refresh()
            except ApiError as e:
                if status == "archived" and e.status == 404:
                    return
                raise e
            if self.status in ["errored", "failed"]:
                raise NodeError(self.status, self)
            if self.status == status:
                return self
            else:
                count += 1
                logging.debug(f"self: {self}, retry count {count}")

        raise ProxiMLException(f"Timeout waiting for {status}")



import re
import json
import logging
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises
from aiohttp import WSMessage, WSMsgType

import proximl.projects.datastores as specimen
from proximl.exceptions import (
    ApiError,
    SpecificationError,
    ProxiMLException,
)

pytestmark = [mark.sdk, mark.unit, mark.projects]


@fixture
def project_datastores(mock_proximl):
    yield specimen.ProjectDatastores(mock_proximl, project_id="1")


@fixture
def project_datastore(mock_proximl):
    yield specimen.ProjectDatastore(
        mock_proximl,
        id="ds-id-1",
        name="datastore 1",
        project_uuid="proj-id-1",
        type="nfs",
        region_uuid="reg-id-1",
    )


class ProjectDatastoresTests:
    @mark.asyncio
    async def test_project_datastores_refresh(self, project_datastores, mock_proximl):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await project_datastores.refresh()
        mock_proximl._query.assert_called_once_with("/project/1/datastores", "PATCH")

    @mark.asyncio
    async def test_project_datastores_list(self, project_datastores, mock_proximl):
        api_response = [
            {
                "project_uuid": "proj-id-1",
                "region_uuid": "reg-id-1",
                "id": "store-id-1",
                "type": "nfs",
                "name": "On-prem NFS",
            },
            {
                "project_uuid": "proj-id-1",
                "region_uuid": "reg-id-2",
                "id": "store-id-2",
                "type": "smb",
                "name": "GCP Samba",
            },
        ]
        mock_proximl._query = AsyncMock(return_value=api_response)
        resp = await project_datastores.list()
        mock_proximl._query.assert_called_once_with(
            "/project/1/datastores", "GET", dict()
        )
        assert len(resp) == 2


class ProjectDatastoreTests:
    def test_project_datastore_properties(self, project_datastore):
        assert isinstance(project_datastore.id, str)
        assert isinstance(project_datastore.name, str)
        assert isinstance(project_datastore.project_uuid, str)
        assert isinstance(project_datastore.type, str)
        assert isinstance(project_datastore.region_uuid, str)

    def test_project_datastore_str(self, project_datastore):
        string = str(project_datastore)
        regex = r"^{.*\"id\": \"" + project_datastore.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_project_datastore_repr(self, project_datastore):
        string = repr(project_datastore)
        regex = (
            r"^ProjectDatastore\( proximl , \*\*{.*'id': '"
            + project_datastore.id
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_project_datastore_bool(self, project_datastore, mock_proximl):
        empty_project_datastore = specimen.ProjectDatastore(mock_proximl)
        assert bool(project_datastore)
        assert not bool(empty_project_datastore)



import re
import json
import logging
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises
from aiohttp import WSMessage, WSMsgType

import proximl.projects.services as specimen
from proximl.exceptions import (
    ApiError,
    SpecificationError,
    ProxiMLException,
)

pytestmark = [mark.sdk, mark.unit, mark.projects]


@fixture
def project_services(mock_proximl):
    yield specimen.ProjectServices(mock_proximl, project_id="1")


@fixture
def project_service(mock_proximl):
    yield specimen.ProjectService(
        mock_proximl,
        id="res-id-1",
        name="service 1",
        project_uuid="proj-id-1",
        region_uuid="reg-id-1",
        public=False,
        hostname="asdf.proximl.cloud",
    )


class ProjectServicesTests:
    @mark.asyncio
    async def test_project_services_refresh(self, project_services, mock_proximl):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await project_services.refresh()
        mock_proximl._query.assert_called_once_with("/project/1/services", "PATCH")

    @mark.asyncio
    async def test_project_services_list(self, project_services, mock_proximl):
        api_response = [
            {
                "project_uuid": "proj-id-1",
                "region_uuid": "reg-id-1",
                "id": "res-id-1",
                "type": "port",
                "name": "On-Prem Service A",
                "resource": "8001",
                "hostname": "service-a.local",
            },
            {
                "project_uuid": "proj-id-1",
                "region_uuid": "reg-id-2",
                "id": "res-id-2",
                "type": "port",
                "name": "Cloud Service B",
                "resource": "8001",
                "hostname": "service-b.local",
            },
        ]
        mock_proximl._query = AsyncMock(return_value=api_response)
        resp = await project_services.list()
        mock_proximl._query.assert_called_once_with(
            "/project/1/services", "GET", dict()
        )
        assert len(resp) == 2


class ProjectServiceTests:
    def test_project_service_properties(self, project_service):
        assert isinstance(project_service.id, str)
        assert isinstance(project_service.name, str)
        assert isinstance(project_service.project_uuid, str)
        assert isinstance(project_service.hostname, str)
        assert isinstance(project_service.public, bool)
        assert isinstance(project_service.region_uuid, str)

    def test_project_service_str(self, project_service):
        string = str(project_service)
        regex = r"^{.*\"id\": \"" + project_service.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_project_service_repr(self, project_service):
        string = repr(project_service)
        regex = (
            r"^ProjectService\( proximl , \*\*{.*'id': '"
            + project_service.id
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_project_service_bool(self, project_service, mock_proximl):
        empty_project_service = specimen.ProjectService(mock_proximl)
        assert bool(project_service)
        assert not bool(empty_project_service)



import json
import logging
import asyncio
import math
from enum import Enum

from proximl.exceptions import (
    ApiError,
    SpecificationError,
    ProxiMLException,
)

class SERVICE_CERT_ALGORITHMS(str, Enum):
    RSA_2048 = 'rsa2048'
    RSA_4096 = 'rsa4096'
    ED25519 = 'ed25519'
    P256 = 'p256'
    P384 ='p384'
    P521='p521'


class Services(object):
    def __init__(self, proximl):
        self.proximl = proximl

    async def get(self, provider_uuid, region_uuid, id, **kwargs):
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/service/{id}",
            "GET",
            kwargs,
        )
        return Service(self.proximl, **resp)

    async def list(self, provider_uuid, region_uuid, **kwargs):
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/service",
            "GET",
            kwargs,
        )
        services = [Service(self.proximl, **service) for service in resp]
        return services

    async def create(
        self,
        provider_uuid,
        region_uuid,
        name,
        type,
        public,
        **kwargs,
    ):
        logging.info(f"Creating Service {name}")
        data = dict(
            name=name,
            type=type,
            public=public,
            **kwargs,
        )
        payload = {k: v for k, v in data.items() if v is not None}
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/service",
            "POST",
            None,
            payload,
        )
        service = Service(self.proximl, **resp)
        logging.info(f"Created Service {name} with id {service.id}")
        return service

    async def remove(self, provider_uuid, region_uuid, id, **kwargs):
        await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/service/{id}",
            "DELETE",
            kwargs,
        )


class Service:
    def __init__(self, proximl,  **kwargs):
        self.proximl = proximl
        self._service = kwargs
        self._id = self._service.get("service_id")
        self._provider_uuid = self._service.get("provider_uuid")
        self._region_uuid = self._service.get("region_uuid")
        self._public = self._service.get("public")
        self._name = self._service.get("name")
        self._type = self._service.get("type")
        self._hostname = self._service.get("custom_hostname") or self._service.get(
            "hostname"
        )
        self._status = self._service.get("status")
        self._port = self._service.get("port")

    @property
    def id(self) -> str:
        return self._id

    @property
    def provider_uuid(self) -> str:
        return self._provider_uuid

    @property
    def region_uuid(self) -> str:
        return self._region_uuid

    @property
    def public(self) -> bool:
        return self._public

    @property
    def name(self) -> str:
        return self._name

    @property
    def hostname(self) -> str:
        return self._hostname

    @property
    def status(self) -> str:
        return self._status

    @property
    def type(self) -> str:
        return self._type

    @property
    def port(self) -> str:
        return self._port

    def __str__(self):
        return json.dumps({k: v for k, v in self._service.items()})

    def __repr__(self):
        return f"Service( proximl , **{self._service.__repr__()})"

    def __bool__(self):
        return bool(self._id)

    async def remove(self):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/service/{self._id}",
            "DELETE",
        )

    async def refresh(self):
        resp = await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/service/{self._id}",
            "GET",
        )
        self.__init__(self.proximl, **resp)
        return self

    async def wait_for(self, status, timeout=300):
        if self.status == status:
            return
        valid_statuses = ["active", "archived"]
        if not status in valid_statuses:
            raise SpecificationError(
                "status",
                f"Invalid wait_for status {status}.  Valid statuses are: {valid_statuses}",
            )
        MAX_TIMEOUT = 24 * 60 * 60
        if timeout > MAX_TIMEOUT:
            raise SpecificationError(
                "timeout",
                f"timeout must be less than {MAX_TIMEOUT} seconds.",
            )

        POLL_INTERVAL_MIN = 5
        POLL_INTERVAL_MAX = 60
        POLL_INTERVAL = max(min(timeout / 60, POLL_INTERVAL_MAX), POLL_INTERVAL_MIN)
        retry_count = math.ceil(timeout / POLL_INTERVAL)
        count = 0
        while count < retry_count:
            await asyncio.sleep(POLL_INTERVAL)
            try:
                await self.refresh()
            except ApiError as e:
                if status == "archived" and e.status == 404:
                    return
                raise e
            if self.status == status:
                return self
            else:
                count += 1
                logging.debug(f"self: {self}, retry count {count}")

        raise ProxiMLException(f"Timeout waiting for {status}")
    
    async def generate_certificate(self, algorithm=SERVICE_CERT_ALGORITHMS.ED25519.value, **kwargs):
        resp = await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/service/{self._id}/certificate",
            "POST",
            kwargs,
            dict(algorithm=algorithm)
        )
        self.__init__(self.proximl, **resp)
        return self
    
    async def sign_client_certificate(self, csr, **kwargs):
        certificate = await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/service/{self._id}/certificate/sign",
            "POST",
            kwargs,
            dict(csr=csr)
        )
        return certificate



import re
import json
import logging
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises
from aiohttp import WSMessage, WSMsgType

import proximl.projects.secrets as specimen
from proximl.exceptions import (
    ApiError,
    SpecificationError,
    ProxiMLException,
)

pytestmark = [mark.sdk, mark.unit, mark.projects]


@fixture
def project_secrets(mock_proximl):
    yield specimen.ProjectSecrets(mock_proximl, project_id="1")


@fixture
def project_service(mock_proximl):
    yield specimen.ProjectSecret(
        mock_proximl,
        project_uuid="proj-id-1",
        name="secret_value",
    )


class ProjectSecretsTests:
    @mark.asyncio
    async def test_project_secrets_list(self, project_secrets, mock_proximl):
        api_response = [
            {
                "project_uuid": "proj-id-1",
                "name": "secret_value",
            },
            {
                "project_uuid": "proj-id-1",
                "name": "secret_value_2",
            },
        ]
        mock_proximl._query = AsyncMock(return_value=api_response)
        resp = await project_secrets.list()
        mock_proximl._query.assert_called_once_with("/project/1/secrets", "GET", dict())
        assert len(resp) == 2

    @mark.asyncio
    async def test_remove_project_secret(
        self,
        project_secrets,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await project_secrets.remove("secret_value")
        mock_proximl._query.assert_called_once_with(
            "/project/1/secret/secret_value", "DELETE", dict()
        )

    @mark.asyncio
    async def test_put_project_secret(self, project_secrets, mock_proximl):
        requested_config = dict(name="secret_value", value="ASKHJSLKF")
        expected_payload = dict(value="ASKHJSLKF")
        api_response = {"project_uuid": "project-id-1", "name": "secret_value"}

        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await project_secrets.put(**requested_config)
        mock_proximl._query.assert_called_once_with(
            "/project/1/secret/secret_value", "PUT", None, expected_payload
        )
        assert response.name == "secret_value"


class ProjectSecretTests:
    def test_project_service_properties(self, project_service):
        assert isinstance(project_service.name, str)
        assert isinstance(project_service.project_uuid, str)

    def test_project_service_str(self, project_service):
        string = str(project_service)
        regex = r"^{.*\"name\": \"" + project_service.name + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_project_service_repr(self, project_service):
        string = repr(project_service)
        regex = (
            r"^ProjectSecret\( proximl , \*\*{.*'name': '"
            + project_service.name
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_project_service_bool(self, project_service, mock_proximl):
        empty_project_service = specimen.ProjectSecret(mock_proximl)
        assert bool(project_service)
        assert not bool(empty_project_service)



import re
import json
import click
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises

pytestmark = [mark.cli, mark.unit, mark.volumes]

from proximl.cli import volume as specimen
from proximl.volumes import Volume


def test_list(runner, mock_my_volumes):
    with patch("proximl.cli.ProxiML", new=AsyncMock) as mock_proximl:
        mock_proximl.volumes = AsyncMock()
        mock_proximl.volumes.list = AsyncMock(return_value=mock_my_volumes)
        result = runner.invoke(specimen, ["list"])
        print(result)
        assert result.exit_code == 0
        mock_proximl.volumes.list.assert_called_once()



import click
import os
import json
import base64
from pathlib import Path
from proximl.cli import pass_config
from proximl.cli.project import project


@project.group()
@pass_config
def datastore(config):
    """proxiML project datastore commands."""
    pass


@datastore.command()
@pass_config
def list(config):
    """List project datastores."""
    data = [
        ["ID", "NAME", "TYPE", "REGION_UUID"],
        [
            "-" * 80,
            "-" * 80,
            "-" * 80,
            "-" * 80,
        ],
    ]
    project = config.proximl.run(
        config.proximl.client.projects.get(config.proximl.client.project)
    )

    datastores = config.proximl.run(project.datastores.list())

    for datastore in datastores:
        data.append(
            [
                datastore.id,
                datastore.name,
                datastore.type,
                datastore.region_uuid,
            ]
        )

    for row in data:
        click.echo(
            "{: >38.36} {: >30.28} {: >15.13} {: >38.36}" "".format(*row),
            file=config.stdout,
        )


@datastore.command()
@pass_config
def refresh(config):
    """
    Refresh project datastore list.
    """
    project = config.proximl.run(config.proximl.client.projects.get_current())

    return config.proximl.run(project.datastores.refresh())



import click
from proximl.cli import cli, pass_config, search_by_id_name
from proximl.cli.cloudbender import cloudbender


@cloudbender.group()
@pass_config
def data_connector(config):
    """proxiML CloudBender data connector commands."""
    pass


@data_connector.command()
@click.option(
    "--provider",
    "-p",
    type=click.STRING,
    required=True,
    help="The provider ID of the region.",
)
@click.option(
    "--region",
    "-r",
    type=click.STRING,
    required=True,
    help="The region ID to list data connectors for.",
)
@pass_config
def list(config, provider, region):
    """List data connectors."""
    data = [
        ["ID", "NAME", "TYPE"],
        [
            "-" * 80,
            "-" * 80,
            "-" * 80,
        ],
    ]

    data_connectors = config.proximl.run(
        config.proximl.client.cloudbender.data_connectors.list(
            provider_uuid=provider, region_uuid=region
        )
    )

    for data_connector in data_connectors:
        data.append(
            [
                data_connector.id,
                data_connector.name,
                data_connector.type,
            ]
        )

    for row in data:
        click.echo(
            "{: >37.36} {: >29.28} {: >9.8}" "".format(*row),
            file=config.stdout,
        )


@data_connector.command()
@click.option(
    "--provider",
    "-p",
    type=click.STRING,
    required=True,
    help="The provider ID of the region.",
)
@click.option(
    "--region",
    "-r",
    type=click.STRING,
    required=True,
    help="The region ID to create the data_connector in.",
)
@click.option(
    "--type",
    "-t",
    type=click.Choice(
        [
            "custom",
        ],
        case_sensitive=False,
    ),
    required=True,
    help="The type of data connector to create.",
)
@click.option(
    "--protocol",
    "-r",
    type=click.STRING,
    help="The transport protocol of the data connector",
)
@click.option(
    "--port-range",
    "-p",
    type=click.STRING,
    help="The port range of the data connector",
)
@click.option(
    "--cidr",
    "-i",
    type=click.STRING,
    help="The IP range to allow in CIDR notation",
)
@click.argument("name", type=click.STRING, required=True)
@pass_config
def create(config, provider, region, type, protocol, port_range, cidr, name):
    """
    Creates a data_connector.
    """
    return config.proximl.run(
        config.proximl.client.cloudbender.data_connectors.create(
            provider_uuid=provider,
            region_uuid=region,
            name=name,
            type=type,
            protocol=protocol,
            port_range=port_range,
            cidr=cidr,
        )
    )


@data_connector.command()
@click.option(
    "--provider",
    "-p",
    type=click.STRING,
    required=True,
    help="The provider ID of the region.",
)
@click.option(
    "--region",
    "-r",
    type=click.STRING,
    required=True,
    help="The region ID to remove the data_connector from.",
)
@click.argument("data_connector", type=click.STRING)
@pass_config
def remove(config, provider, region, data_connector):
    """
    Remove a data_connector.

    DATASTORE may be specified by name or ID, but ID is preferred.
    """
    data_connectors = config.proximl.run(
        config.proximl.client.cloudbender.data_connectors.list(
            provider_uuid=provider, region_uuid=region
        )
    )

    found = search_by_id_name(data_connector, data_connectors)
    if None is found:
        raise click.UsageError("Cannot find specified data_connector.")

    return config.proximl.run(found.remove())



import re
import sys
import asyncio
from pytest import mark, fixture

pytestmark = [mark.sdk, mark.integration, mark.cloudbender, mark.regions]


@mark.create
@mark.asyncio
@mark.xdist_group("cloudbender_resources")
class GetRegionTests:
    async def test_get_regions(self, proximl, provider, region):
        regions = await proximl.cloudbender.regions.list(provider_uuid=provider.id)
        assert len(regions) > 0

    async def test_get_region(self, proximl, provider, region):
        response = await proximl.cloudbender.regions.get(provider.id, region.id)
        assert response.id == region.id

    async def test_region_properties(self, provider, region):
        assert isinstance(region.id, str)
        assert isinstance(region.provider_uuid, str)
        assert isinstance(region.type, str)
        assert region.type == "test"
        assert region.provider_uuid == provider.id

    async def test_region_str(self, region):
        string = str(region)
        regex = r"^{.*\"region_uuid\": \"" + region.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    async def test_region_repr(self, region):
        string = repr(region)
        regex = (
            r"^Region\( proximl , \*\*{.*'region_uuid': '"
            + region.id
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)



##
##  Source: https://github.com/capless/warrant.git
##          https://github.com/capless/warrant/blob/master/warrant/aws_srp.py
##
#                                 Apache License
#                            Version 2.0, January 2004
#                         http://www.apache.org/licenses/

#    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION

#    1. Definitions.

#       "License" shall mean the terms and conditions for use, reproduction,
#       and distribution as defined by Sections 1 through 9 of this document.

#       "Licensor" shall mean the copyright owner or entity authorized by
#       the copyright owner that is granting the License.

#       "Legal Entity" shall mean the union of the acting entity and all
#       other entities that control, are controlled by, or are under common
#       control with that entity. For the purposes of this definition,
#       "control" means (i) the power, direct or indirect, to cause the
#       direction or management of such entity, whether by contract or
#       otherwise, or (ii) ownership of fifty percent (50%) or more of the
#       outstanding shares, or (iii) beneficial ownership of such entity.

#       "You" (or "Your") shall mean an individual or Legal Entity
#       exercising permissions granted by this License.

#       "Source" form shall mean the preferred form for making modifications,
#       including but not limited to software source code, documentation
#       source, and configuration files.

#       "Object" form shall mean any form resulting from mechanical
#       transformation or translation of a Source form, including but
#       not limited to compiled object code, generated documentation,
#       and conversions to other media types.

#       "Work" shall mean the work of authorship, whether in Source or
#       Object form, made available under the License, as indicated by a
#       copyright notice that is included in or attached to the work
#       (an example is provided in the Appendix below).

#       "Derivative Works" shall mean any work, whether in Source or Object
#       form, that is based on (or derived from) the Work and for which the
#       editorial revisions, annotations, elaborations, or other modifications
#       represent, as a whole, an original work of authorship. For the purposes
#       of this License, Derivative Works shall not include works that remain
#       separable from, or merely link (or bind by name) to the interfaces of,
#       the Work and Derivative Works thereof.

#       "Contribution" shall mean any work of authorship, including
#       the original version of the Work and any modifications or additions
#       to that Work or Derivative Works thereof, that is intentionally
#       submitted to Licensor for inclusion in the Work by the copyright owner
#       or by an individual or Legal Entity authorized to submit on behalf of
#       the copyright owner. For the purposes of this definition, "submitted"
#       means any form of electronic, verbal, or written communication sent
#       to the Licensor or its representatives, including but not limited to
#       communication on electronic mailing lists, source code control systems,
#       and issue tracking systems that are managed by, or on behalf of, the
#       Licensor for the purpose of discussing and improving the Work, but
#       excluding communication that is conspicuously marked or otherwise
#       designated in writing by the copyright owner as "Not a Contribution."

#       "Contributor" shall mean Licensor and any individual or Legal Entity
#       on behalf of whom a Contribution has been received by Licensor and
#       subsequently incorporated within the Work.

#    2. Grant of Copyright License. Subject to the terms and conditions of
#       this License, each Contributor hereby grants to You a perpetual,
#       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
#       copyright license to reproduce, prepare Derivative Works of,
#       publicly display, publicly perform, sublicense, and distribute the
#       Work and such Derivative Works in Source or Object form.

#    3. Grant of Patent License. Subject to the terms and conditions of
#       this License, each Contributor hereby grants to You a perpetual,
#       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
#       (except as stated in this section) patent license to make, have made,
#       use, offer to sell, sell, import, and otherwise transfer the Work,
#       where such license applies only to those patent claims licensable
#       by such Contributor that are necessarily infringed by their
#       Contribution(s) alone or by combination of their Contribution(s)
#       with the Work to which such Contribution(s) was submitted. If You
#       institute patent litigation against any entity (including a
#       cross-claim or counterclaim in a lawsuit) alleging that the Work
#       or a Contribution incorporated within the Work constitutes direct
#       or contributory patent infringement, then any patent licenses
#       granted to You under this License for that Work shall terminate
#       as of the date such litigation is filed.

#    4. Redistribution. You may reproduce and distribute copies of the
#       Work or Derivative Works thereof in any medium, with or without
#       modifications, and in Source or Object form, provided that You
#       meet the following conditions:

#       (a) You must give any other recipients of the Work or
#           Derivative Works a copy of this License; and

#       (b) You must cause any modified files to carry prominent notices
#           stating that You changed the files; and

#       (c) You must retain, in the Source form of any Derivative Works
#           that You distribute, all copyright, patent, trademark, and
#           attribution notices from the Source form of the Work,
#           excluding those notices that do not pertain to any part of
#           the Derivative Works; and

#       (d) If the Work includes a "NOTICE" text file as part of its
#           distribution, then any Derivative Works that You distribute must
#           include a readable copy of the attribution notices contained
#           within such NOTICE file, excluding those notices that do not
#           pertain to any part of the Derivative Works, in at least one
#           of the following places: within a NOTICE text file distributed
#           as part of the Derivative Works; within the Source form or
#           documentation, if provided along with the Derivative Works; or,
#           within a display generated by the Derivative Works, if and
#           wherever such third-party notices normally appear. The contents
#           of the NOTICE file are for informational purposes only and
#           do not modify the License. You may add Your own attribution
#           notices within Derivative Works that You distribute, alongside
#           or as an addendum to the NOTICE text from the Work, provided
#           that such additional attribution notices cannot be construed
#           as modifying the License.

#       You may add Your own copyright statement to Your modifications and
#       may provide additional or different license terms and conditions
#       for use, reproduction, or distribution of Your modifications, or
#       for any such Derivative Works as a whole, provided Your use,
#       reproduction, and distribution of the Work otherwise complies with
#       the conditions stated in this License.

#    5. Submission of Contributions. Unless You explicitly state otherwise,
#       any Contribution intentionally submitted for inclusion in the Work
#       by You to the Licensor shall be under the terms and conditions of
#       this License, without any additional terms or conditions.
#       Notwithstanding the above, nothing herein shall supersede or modify
#       the terms of any separate license agreement you may have executed
#       with Licensor regarding such Contributions.

#    6. Trademarks. This License does not grant permission to use the trade
#       names, trademarks, service marks, or product names of the Licensor,
#       except as required for reasonable and customary use in describing the
#       origin of the Work and reproducing the content of the NOTICE file.

#    7. Disclaimer of Warranty. Unless required by applicable law or
#       agreed to in writing, Licensor provides the Work (and each
#       Contributor provides its Contributions) on an "AS IS" BASIS,
#       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
#       implied, including, without limitation, any warranties or conditions
#       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
#       PARTICULAR PURPOSE. You are solely responsible for determining the
#       appropriateness of using or redistributing the Work and assume any
#       risks associated with Your exercise of permissions under this License.

#    8. Limitation of Liability. In no event and under no legal theory,
#       whether in tort (including negligence), contract, or otherwise,
#       unless required by applicable law (such as deliberate and grossly
#       negligent acts) or agreed to in writing, shall any Contributor be
#       liable to You for damages, including any direct, indirect, special,
#       incidental, or consequential damages of any character arising as a
#       result of this License or out of the use or inability to use the
#       Work (including but not limited to damages for loss of goodwill,
#       work stoppage, computer failure or malfunction, or any and all
#       other commercial damages or losses), even if such Contributor
#       has been advised of the possibility of such damages.

#    9. Accepting Warranty or Additional Liability. While redistributing
#       the Work or Derivative Works thereof, You may choose to offer,
#       and charge a fee for, acceptance of support, warranty, indemnity,
#       or other liability obligations and/or rights consistent with this
#       License. However, in accepting such obligations, You may act only
#       on Your own behalf and on Your sole responsibility, not on behalf
#       of any other Contributor, and only if You agree to indemnify,
#       defend, and hold each Contributor harmless for any liability
#       incurred by, or claims asserted against, such Contributor by reason
#       of your accepting any such warranty or additional liability.

#    END OF TERMS AND CONDITIONS
##
##
import base64
import binascii
import datetime
import hashlib
import hmac
import re
import json
import requests
import logging
import time
from datetime import datetime

import boto3
import os
import six
from jose import jwt

from proximl.exceptions import ProxiMLException

# https://github.com/aws/amazon-cognito-identity-js/blob/master/src/AuthenticationHelper.js#L22
n_hex = (
    "FFFFFFFFFFFFFFFFC90FDAA22168C234C4C6628B80DC1CD1"
    + "29024E088A67CC74020BBEA63B139B22514A08798E3404DD"
    + "EF9519B3CD3A431B302B0A6DF25F14374FE1356D6D51C245"
    + "E485B576625E7EC6F44C42E9A637ED6B0BFF5CB6F406B7ED"
    + "EE386BFB5A899FA5AE9F24117C4B1FE649286651ECE45B3D"
    + "C2007CB8A163BF0598DA48361C55D39A69163FA8FD24CF5F"
    + "83655D23DCA3AD961C62F356208552BB9ED529077096966D"
    + "670C354E4ABC9804F1746C08CA18217C32905E462E36CE3B"
    + "E39E772C180E86039B2783A2EC07A28FB5C55DF06F4C52C9"
    + "DE2BCBF6955817183995497CEA956AE515D2261898FA0510"
    + "15728E5A8AAAC42DAD33170D04507A33A85521ABDF1CBA64"
    + "ECFB850458DBEF0A8AEA71575D060C7DB3970F85A6E1E4C7"
    + "ABF5AE8CDB0933D71E8C94E04A25619DCEE3D2261AD2EE6B"
    + "F12FFA06D98A0864D87602733EC86A64521F2B18177B200C"
    + "BBE117577A615D6C770988C0BAD946E208E24FA074E5AB31"
    + "43DB5BFCE0FD108E4B82D120A93AD2CAFFFFFFFFFFFFFFFF"
)
# https://github.com/aws/amazon-cognito-identity-js/blob/master/src/AuthenticationHelper.js#L49
g_hex = "2"
info_bits = bytearray("Caldera Derived Key", "utf-8")


def hash_sha256(buf):
    """AuthenticationHelper.hash"""
    a = hashlib.sha256(buf).hexdigest()
    return (64 - len(a)) * "0" + a


def hex_hash(hex_string):
    return hash_sha256(bytearray.fromhex(hex_string))


def hex_to_long(hex_string):
    return int(hex_string, 16)


def long_to_hex(long_num):
    return "%x" % long_num


def get_random(nbytes):
    random_hex = binascii.hexlify(os.urandom(nbytes))
    return hex_to_long(random_hex)


def pad_hex(long_int):
    """
    Converts a Long integer (or hex string) to hex format padded with zeroes for hashing
    :param {Long integer|String} long_int Number or string to pad.
    :return {String} Padded hex string.
    """
    if not isinstance(long_int, six.string_types):
        hash_str = long_to_hex(long_int)
    else:
        hash_str = long_int
    if len(hash_str) % 2 == 1:
        hash_str = "0%s" % hash_str
    elif hash_str[0] in "89ABCDEFabcdef":
        hash_str = "00%s" % hash_str
    return hash_str


def compute_hkdf(ikm, salt):
    """
    Standard hkdf algorithm
    :param {Buffer} ikm Input key material.
    :param {Buffer} salt Salt value.
    :return {Buffer} Strong key material.
    @private
    """
    prk = hmac.new(salt, ikm, hashlib.sha256).digest()
    info_bits_update = info_bits + bytearray(chr(1), "utf-8")
    hmac_hash = hmac.new(prk, info_bits_update, hashlib.sha256).digest()
    return hmac_hash[:16]


def calculate_u(big_a, big_b):
    """
    Calculate the client's value U which is the hash of A and B
    :param {Long integer} big_a Large A value.
    :param {Long integer} big_b Server B value.
    :return {Long integer} Computed U value.
    """
    u_hex_hash = hex_hash(pad_hex(big_a) + pad_hex(big_b))
    return hex_to_long(u_hex_hash)


class AWSSRP(object):
    NEW_PASSWORD_REQUIRED_CHALLENGE = "NEW_PASSWORD_REQUIRED"
    PASSWORD_VERIFIER_CHALLENGE = "PASSWORD_VERIFIER"

    def __init__(
        self,
        username,
        password,
        pool_id,
        client_id,
        pool_region=None,
        client=None,
        client_secret=None,
    ):
        if pool_region is not None and client is not None:
            raise ValueError(
                "pool_region and client should not both be specified "
                "(region should be passed to the boto3 client instead)"
            )

        self.username = username
        self.password = password
        self.pool_id = pool_id
        self.client_id = client_id
        self.client_secret = client_secret
        self.client = (
            client
            if client
            else boto3.client("cognito-idp", region_name=pool_region)
        )
        self.big_n = hex_to_long(n_hex)
        self.g = hex_to_long(g_hex)
        self.k = hex_to_long(hex_hash("00" + n_hex + "0" + g_hex))
        self.small_a_value = self.generate_random_small_a()
        self.large_a_value = self.calculate_a()

    def generate_random_small_a(self):
        """
        helper function to generate a random big integer
        :return {Long integer} a random value.
        """
        random_long_int = get_random(128)
        return random_long_int % self.big_n

    def calculate_a(self):
        """
        Calculate the client's public value A = g^a%N
        with the generated random number a
        :param {Long integer} a Randomly generated small A.
        :return {Long integer} Computed large A.
        """
        big_a = pow(self.g, self.small_a_value, self.big_n)
        # safety check
        if (big_a % self.big_n) == 0:
            raise ValueError("Safety check for A failed")
        return big_a

    def get_password_authentication_key(
        self, username, password, server_b_value, salt
    ):
        """
        Calculates the final hkdf based on computed S value, and computed U value and the key
        :param {String} username Username.
        :param {String} password Password.
        :param {Long integer} server_b_value Server B value.
        :param {Long integer} salt Generated salt.
        :return {Buffer} Computed HKDF value.
        """
        u_value = calculate_u(self.large_a_value, server_b_value)
        if u_value == 0:
            raise ValueError("U cannot be zero.")
        username_password = "%s%s:%s" % (
            self.pool_id.split("_")[1],
            username,
            password,
        )
        username_password_hash = hash_sha256(username_password.encode("utf-8"))

        x_value = hex_to_long(hex_hash(pad_hex(salt) + username_password_hash))
        g_mod_pow_xn = pow(self.g, x_value, self.big_n)
        int_value2 = server_b_value - self.k * g_mod_pow_xn
        s_value = pow(
            int_value2, self.small_a_value + u_value * x_value, self.big_n
        )
        hkdf = compute_hkdf(
            bytearray.fromhex(pad_hex(s_value)),
            bytearray.fromhex(pad_hex(long_to_hex(u_value))),
        )
        return hkdf

    def get_auth_params(self):
        auth_params = {
            "USERNAME": self.username,
            "SRP_A": long_to_hex(self.large_a_value),
        }
        if self.client_secret is not None:
            auth_params.update(
                {
                    "SECRET_HASH": self.get_secret_hash(
                        self.username, self.client_id, self.client_secret
                    )
                }
            )
        return auth_params

    @staticmethod
    def get_secret_hash(username, client_id, client_secret):
        message = bytearray(username + client_id, "utf-8")
        hmac_obj = hmac.new(
            bytearray(client_secret, "utf-8"), message, hashlib.sha256
        )
        return base64.standard_b64encode(hmac_obj.digest()).decode("utf-8")

    def process_challenge(self, challenge_parameters):
        user_id_for_srp = challenge_parameters["USER_ID_FOR_SRP"]
        salt_hex = challenge_parameters["SALT"]
        srp_b_hex = challenge_parameters["SRP_B"]
        secret_block_b64 = challenge_parameters["SECRET_BLOCK"]
        # re strips leading zero from a day number (required by AWS Cognito)
        timestamp = re.sub(
            r" 0(\d) ",
            r" \1 ",
            datetime.utcnow().strftime("%a %b %d %H:%M:%S UTC %Y"),
        )
        hkdf = self.get_password_authentication_key(
            user_id_for_srp, self.password, hex_to_long(srp_b_hex), salt_hex
        )
        secret_block_bytes = base64.standard_b64decode(secret_block_b64)
        msg = (
            bytearray(self.pool_id.split("_")[1], "utf-8")
            + bytearray(user_id_for_srp, "utf-8")
            + bytearray(secret_block_bytes)
            + bytearray(timestamp, "utf-8")
        )
        hmac_obj = hmac.new(hkdf, msg, digestmod=hashlib.sha256)
        signature_string = base64.standard_b64encode(hmac_obj.digest())
        response = {
            "TIMESTAMP": timestamp,
            "USERNAME": user_id_for_srp,
            "PASSWORD_CLAIM_SECRET_BLOCK": secret_block_b64,
            "PASSWORD_CLAIM_SIGNATURE": signature_string.decode("utf-8"),
        }
        if self.client_secret is not None:
            response.update(
                {
                    "SECRET_HASH": self.get_secret_hash(
                        self.username, self.client_id, self.client_secret
                    )
                }
            )
        return response

    def authenticate_user(self, client=None):
        boto_client = self.client or client
        auth_params = self.get_auth_params()
        response = boto_client.initiate_auth(
            AuthFlow="USER_SRP_AUTH",
            AuthParameters=auth_params,
            ClientId=self.client_id,
        )
        if response["ChallengeName"] == self.PASSWORD_VERIFIER_CHALLENGE:
            challenge_response = self.process_challenge(
                response["ChallengeParameters"]
            )
            tokens = boto_client.respond_to_auth_challenge(
                ClientId=self.client_id,
                ChallengeName=self.PASSWORD_VERIFIER_CHALLENGE,
                ChallengeResponses=challenge_response,
            )

            if (
                tokens.get("ChallengeName")
                == self.NEW_PASSWORD_REQUIRED_CHALLENGE
            ):
                raise Exception("Change password before authenticating")

            return tokens
        else:
            raise NotImplementedError(
                "The %s challenge is not supported" % response["ChallengeName"]
            )

    def set_new_password_challenge(self, new_password, client=None):
        boto_client = self.client or client
        auth_params = self.get_auth_params()
        response = boto_client.initiate_auth(
            AuthFlow="USER_SRP_AUTH",
            AuthParameters=auth_params,
            ClientId=self.client_id,
        )
        if response["ChallengeName"] == self.PASSWORD_VERIFIER_CHALLENGE:
            challenge_response = self.process_challenge(
                response["ChallengeParameters"]
            )
            tokens = boto_client.respond_to_auth_challenge(
                ClientId=self.client_id,
                ChallengeName=self.PASSWORD_VERIFIER_CHALLENGE,
                ChallengeResponses=challenge_response,
            )

            if tokens["ChallengeName"] == self.NEW_PASSWORD_REQUIRED_CHALLENGE:
                challenge_response = {
                    "USERNAME": auth_params["USERNAME"],
                    "NEW_PASSWORD": new_password,
                }
                new_password_response = boto_client.respond_to_auth_challenge(
                    ClientId=self.client_id,
                    ChallengeName=self.NEW_PASSWORD_REQUIRED_CHALLENGE,
                    Session=tokens["Session"],
                    ChallengeResponses=challenge_response,
                )
                return new_password_response
            return tokens
        else:
            raise NotImplementedError(
                "The %s challenge is not supported" % response["ChallengeName"]
            )


class Auth(object):
    def __init__(self, config_dir, domain_suffix="proximl.ai", **kwargs):
        try:
            with open(f"{config_dir}/environment.json", "r") as file:
                env_str = file.read().replace("\n", "")
            env = json.loads(env_str)
        except:
            env = dict()

        auth_defaults = requests.get(
            "https://app.{}/.well-known/auth-config.json".format(domain_suffix)
        ).json()

        self.region = (
            kwargs.get("region")
            or os.environ.get("PROXIML_REGION")
            or env.get("region")
            or auth_defaults.get("region")
        )
        self.client_id = (
            kwargs.get("client_id")
            or os.environ.get("PROXIML_CLIENT_ID")
            or env.get("client_id")
            or auth_defaults.get("userPoolSDKClientId")
        )
        self.pool_id = (
            kwargs.get("pool_id")
            or os.environ.get("PROXIML_POOL_ID")
            or env.get("pool_id")
            or auth_defaults.get("userPoolId")
        )

        try:
            with open(f"{config_dir}/credentials.json", "r") as file:
                key_str = file.read().replace("\n", "")
            keys = json.loads(key_str)
        except:
            keys = dict()

        self.username = (
            kwargs.get("user")
            or os.environ.get("PROXIML_USER")
            or keys.get("user")
        )
        self.password = (
            kwargs.get("key")
            or os.environ.get("PROXIML_KEY")
            or keys.get("key")
        )
        if not self.username or not self.password:
            raise ProxiMLException("proxiML credentials not found.")
        self.client = boto3.client("cognito-idp", region_name=self.region)
        self.id_token = None
        self.access_token = None
        self.refresh_token = None
        self.expires = 0

    def get_keys(self):
        pool_jwk = requests.get(
            "https://cognito-idp.{}.amazonaws.com/{}/.well-known/jwks.json".format(
                self.region, self.pool_id
            )
        ).json()
        return pool_jwk

    def get_key(self, kid):
        keys = self.get_keys().get("keys")
        key = list(filter(lambda x: x.get("kid") == kid, keys))
        return key[0]

    def verify_token(self, token, id_name):
        kid = jwt.get_unverified_header(token).get("kid")
        unverified_claims = jwt.get_unverified_claims(token)
        hmac_key = self.get_key(kid)
        try:
            verified = jwt.decode(
                token,
                hmac_key,
                algorithms=["RS256"],
                audience=unverified_claims.get("aud"),
                issuer=unverified_claims.get("iss"),
            )
        except Exception:
            return False
        return verified

    def get_new_tokens(self):
        aws = AWSSRP(
            username=self.username,
            password=self.password,
            pool_id=self.pool_id,
            client_id=self.client_id,
            client=self.client,
        )
        tokens = aws.authenticate_user()
        refresh_token = tokens["AuthenticationResult"]["RefreshToken"]
        id_verify = self.verify_token(
            tokens["AuthenticationResult"]["IdToken"], "id_token"
        )
        logging.debug(f"ID Token Verification: {id_verify}")
        if id_verify:
            id_token = tokens["AuthenticationResult"]["IdToken"]
        access_verify = self.verify_token(
            tokens["AuthenticationResult"]["AccessToken"], "access_token"
        )
        logging.debug(f"Access Token Verification: {access_verify}")
        if access_verify:
            access_token = tokens["AuthenticationResult"]["AccessToken"]

        self.id_token = id_token
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.expires = (
            id_verify.get("exp") - 300
        )  ## prevent just about to expire tokens from being used

    def get_tokens(self):
        logging.debug(f"Token expires: {self.expires}")
        logging.debug(f"Token is expired: {self.expires < time.time()}")
        if not self.id_token:
            self.get_new_tokens()
        elif self.expires < time.time():
            self.get_new_tokens()
        logging.debug(f"New token expires: {self.expires}")

        return dict(
            id_token=self.id_token,
            access_token=self.access_token,
            refresh_token=self.refresh_token,
            expires=self.expires,
        )



import click
import os
import json
import base64
from pathlib import Path
from proximl.cli import pass_config
from proximl.cli.project import project


@project.group()
@pass_config
def service(config):
    """proxiML project service commands."""
    pass


@service.command()
@pass_config
def list(config):
    """List project services."""
    data = [
        ["ID", "NAME", "TYPE", "REGION_UUID"],
        [
            "-" * 80,
            "-" * 80,
            "-" * 80,
            "-" * 80,
        ],
    ]
    project = config.proximl.run(
        config.proximl.client.projects.get(config.proximl.client.project)
    )

    services = config.proximl.run(project.services.list())

    for service in services:
        data.append(
            [
                service.id,
                service.name,
                service.hostname,
                service.region_uuid,
            ]
        )

    for row in data:
        click.echo(
            "{: >38.36} {: >30.28} {: >15.13} {: >38.36}" "".format(*row),
            file=config.stdout,
        )


@service.command()
@pass_config
def refresh(config):
    """
    Refresh project service list.
    """
    project = config.proximl.run(config.proximl.client.projects.get_current())

    return config.proximl.run(project.services.refresh())



import json
import logging
import asyncio
import math

from proximl.exceptions import (
    ApiError,
    SpecificationError,
    ProxiMLException,
    RegionError,
)


class Regions(object):
    def __init__(self, proximl):
        self.proximl = proximl

    async def get(self, provider_uuid, id, **kwargs):
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{id}", "GET", kwargs
        )
        return Region(self.proximl, **resp)

    async def list(self, provider_uuid, **kwargs):
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region", "GET", kwargs
        )
        regions = [Region(self.proximl, **region) for region in resp]
        return regions

    async def create(self, provider_uuid, name, public, storage, **kwargs):
        logging.info(f"Creating Region {name}")
        data = dict(name=name, public=public, storage=storage, **kwargs)
        payload = payload = {k: v for k, v in data.items() if v is not None}
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region", "POST", None, payload
        )
        region = Region(self.proximl, **resp)
        logging.info(f"Created Region {name} with id {region.id}")
        return region

    async def remove(self, provider_uuid, id, **kwargs):
        await self.proximl._query(
            f"/provider/{provider_uuid}/region/{id}", "DELETE", kwargs
        )


class Region:
    def __init__(self, proximl, **kwargs):
        self.proximl = proximl
        self._region = kwargs
        self._id = self._region.get("region_uuid")
        self._provider_uuid = self._region.get("provider_uuid")
        self._type = self._region.get("provider_type")
        self._name = self._region.get("name")
        self._status = self._region.get("status")

    @property
    def id(self) -> str:
        return self._id

    @property
    def provider_uuid(self) -> str:
        return self._provider_uuid

    @property
    def type(self) -> str:
        return self._type

    @property
    def name(self) -> str:
        return self._name

    @property
    def status(self) -> str:
        return self._status

    def __str__(self):
        return json.dumps({k: v for k, v in self._region.items()})

    def __repr__(self):
        return f"Region( proximl , **{self._region.__repr__()})"

    def __bool__(self):
        return bool(self._id)

    async def remove(self):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._id}", "DELETE"
        )

    async def refresh(self):
        resp = await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._id}",
            "GET",
        )
        self.__init__(self.proximl, **resp)
        return self

    async def add_dataset(self, project_uuid, dataset_uuid, **kwargs):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._id}/dataset",
            "POST",
            None,
            dict(project_uuid=project_uuid, dataset_uuid=dataset_uuid),
        )

    async def add_model(self, project_uuid, model_uuid, **kwargs):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._id}/model",
            "POST",
            None,
            dict(project_uuid=project_uuid, model_uuid=model_uuid),
        )

    async def add_checkpoint(self, project_uuid, checkpoint_uuid, **kwargs):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._id}/checkpoint",
            "POST",
            None,
            dict(project_uuid=project_uuid, checkpoint_uuid=checkpoint_uuid),
        )

    async def wait_for(self, status, timeout=300):
        if self.status == status:
            return
        valid_statuses = ["healthy", "offline", "archived"]
        if not status in valid_statuses:
            raise SpecificationError(
                "status",
                f"Invalid wait_for status {status}.  Valid statuses are: {valid_statuses}",
            )
        MAX_TIMEOUT = 24 * 60 * 60
        if timeout > MAX_TIMEOUT:
            raise SpecificationError(
                "timeout",
                f"timeout must be less than {MAX_TIMEOUT} seconds.",
            )

        POLL_INTERVAL_MIN = 5
        POLL_INTERVAL_MAX = 60
        POLL_INTERVAL = max(min(timeout / 60, POLL_INTERVAL_MAX), POLL_INTERVAL_MIN)
        retry_count = math.ceil(timeout / POLL_INTERVAL)
        count = 0
        while count < retry_count:
            await asyncio.sleep(POLL_INTERVAL)
            try:
                await self.refresh()
            except ApiError as e:
                if status == "archived" and e.status == 404:
                    return
                raise e
            if self.status in ["errored", "failed"]:
                raise RegionError(self.status, self)
            if self.status == status:
                return self
            else:
                count += 1
                logging.debug(f"self: {self}, retry count {count}")

        raise ProxiMLException(f"Timeout waiting for {status}")



import re
import json
import logging
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises
from aiohttp import WSMessage, WSMsgType

import proximl.projects as specimen
from proximl.exceptions import (
    ApiError,
    SpecificationError,
    ProxiMLException,
)

pytestmark = [mark.sdk, mark.unit, mark.projects]


@fixture
def projects(mock_proximl):
    yield specimen.Projects(mock_proximl)


@fixture
def project(mock_proximl):
    yield specimen.Project(
        mock_proximl,
        id="1",
        name="My Mock Project",
        owner=True,
        owner_name="Me",
        created_name="Me",
        job_all=True,
        dataset_all=True,
        model_all=True,
        createdAt="2020-12-31T23:59:59.000Z",
    )


class ProjectsTests:
    @mark.asyncio
    async def test_get_project(
        self,
        projects,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await projects.get("1234")
        mock_proximl._query.assert_called_once_with("/project/1234", "GET", dict())

    @mark.asyncio
    async def test_list_projects(
        self,
        projects,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await projects.list()
        mock_proximl._query.assert_called_once_with("/project", "GET", dict())

    @mark.asyncio
    async def test_remove_project(
        self,
        projects,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await projects.remove("4567")
        mock_proximl._query.assert_called_once_with("/project/4567", "DELETE", dict())

    @mark.asyncio
    async def test_create_project_simple(self, projects, mock_proximl):
        requested_config = dict(
            name="new project",
        )
        expected_payload = dict(name="new project")
        api_response = {
            "id": "project-id-1",
            "name": "new project",
            "owner": True,
            "owner_name": "Me",
            "created_name": "Me",
            "job_all": True,
            "dataset_all": True,
            "model_all": True,
            "createdAt": "2020-12-31T23:59:59.000Z",
        }

        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await projects.create(**requested_config)
        mock_proximl._query.assert_called_once_with(
            "/project", "POST", None, expected_payload
        )
        assert response.id == "project-id-1"


class ProjectTests:
    def test_project_properties(self, project):
        assert isinstance(project.id, str)
        assert isinstance(project.name, str)
        assert isinstance(project.owner_name, str)
        assert isinstance(project.is_owner, bool)

    def test_project_str(self, project):
        string = str(project)
        regex = r"^{.*\"id\": \"" + project.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_project_repr(self, project):
        string = repr(project)
        regex = r"^Project\( proximl , \*\*{.*'id': '" + project.id + r"'.*}\)$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_project_bool(self, project, mock_proximl):
        empty_project = specimen.Project(mock_proximl)
        assert bool(project)
        assert not bool(empty_project)

    @mark.asyncio
    async def test_project_remove(self, project, mock_proximl):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await project.remove()
        mock_proximl._query.assert_called_once_with("/project/1", "DELETE")



import re
import json
import logging
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises
from aiohttp import WSMessage, WSMsgType

import proximl.projects.credentials as specimen
from proximl.exceptions import (
    ApiError,
    SpecificationError,
    ProxiMLException,
)

pytestmark = [mark.sdk, mark.unit, mark.projects]


@fixture
def project_credentials(mock_proximl):
    yield specimen.ProjectCredentials(mock_proximl, project_id="1")


@fixture
def project_credential(mock_proximl):
    yield specimen.ProjectCredential(
        mock_proximl, project_uuid="proj-id-1", type="aws", key_id="AIYHGFSDLK"
    )


class ProjectCredentialsTests:
    @mark.asyncio
    async def test_project_credentials_list(self, project_credentials, mock_proximl):
        api_response = [
            {"project_uuid": "proj-id-1", "type": "aws", "key_id": "AIYHGFSDLK"},
            {"project_uuid": "proj-id-1", "type": "gcp", "key_id": "credentials.json"},
        ]
        mock_proximl._query = AsyncMock(return_value=api_response)
        resp = await project_credentials.list()
        mock_proximl._query.assert_called_once_with(
            "/project/1/credentials", "GET", dict()
        )
        assert len(resp) == 2

    @mark.asyncio
    async def test_remove_project_credential(
        self,
        project_credentials,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await project_credentials.remove("aws")
        mock_proximl._query.assert_called_once_with(
            "/project/1/credential/aws", "DELETE", dict()
        )

    @mark.asyncio
    async def test_put_project_credential(self, project_credentials, mock_proximl):
        requested_config = dict(type="aws", key_id="AIUDHADA", secret="ASKHJSLKF")
        expected_payload = dict(key_id="AIUDHADA", secret="ASKHJSLKF")
        api_response = {
            "project_uuid": "project-id-1",
            "type": "aws",
            "key_id": "AIUDHADA",
        }

        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await project_credentials.put(**requested_config)
        mock_proximl._query.assert_called_once_with(
            "/project/1/credential/aws", "PUT", None, expected_payload
        )
        assert response.type == "aws"


class ProjectCredentialTests:
    def test_project_credential_properties(self, project_credential):
        assert isinstance(project_credential.type, str)
        assert isinstance(project_credential.key_id, str)
        assert isinstance(project_credential.project_uuid, str)

    def test_project_credential_str(self, project_credential):
        string = str(project_credential)
        regex = r"^{.*\"type\": \"" + project_credential.type + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_project_credential_repr(self, project_credential):
        string = repr(project_credential)
        regex = (
            r"^ProjectCredential\( proximl , \*\*{.*'type': '"
            + project_credential.type
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_project_credential_bool(self, project_credential, mock_proximl):
        empty_project_credential = specimen.ProjectCredential(mock_proximl)
        assert bool(project_credential)
        assert not bool(empty_project_credential)



import re
import json
import logging
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises
from aiohttp import WSMessage, WSMsgType

import proximl.cloudbender.data_connectors as specimen
from proximl.exceptions import (
    ApiError,
    SpecificationError,
    ProxiMLException,
)

pytestmark = [mark.sdk, mark.unit, mark.cloudbender, mark.data_connectors]


@fixture
def data_connectors(mock_proximl):
    yield specimen.DataConnectors(mock_proximl)


@fixture
def data_connector(mock_proximl):
    yield specimen.DataConnector(
        mock_proximl,
        provider_uuid="1",
        region_uuid="a",
        connector_id="x",
        name="On-Prem Data Connector",
        type="custom",
        cidr="192.168.0.50/32",
        port="443",
        protocol="tcp",
    )


class RegionsTests:
    @mark.asyncio
    async def test_get_data_connector(
        self,
        data_connectors,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await data_connectors.get("1234", "5687", "91011")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/5687/data_connector/91011", "GET", {}
        )

    @mark.asyncio
    async def test_list_data_connectors(
        self,
        data_connectors,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await data_connectors.list("1234", "5687")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/5687/data_connector", "GET", {}
        )

    @mark.asyncio
    async def test_remove_data_connector(
        self,
        data_connectors,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await data_connectors.remove("1234", "4567", "8910")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/4567/data_connector/8910", "DELETE", {}
        )

    @mark.asyncio
    async def test_create_data_connector(self, data_connectors, mock_proximl):
        requested_config = dict(
            provider_uuid="provider-id-1",
            region_uuid="region-id-1",
            name="On-Prem DataConnector",
            type="custom",
            cidr="192.168.0.50/32",
            port="443",
            protocol="tcp",
        )
        expected_payload = dict(
            name="On-Prem DataConnector",
            type="custom",
            cidr="192.168.0.50/32",
            port="443",
            protocol="tcp",
        )
        api_response = {
            "provider_uuid": "provider-id-1",
            "region_uuid": "region-id-1",
            "connector_id": "connector-id-1",
            "name": "On-Prem DataConnector",
            "type": "custom",
            "cidr": "192.168.0.50/32",
            "port": "443",
            "protocol": "tcp",
            "createdAt": "2020-12-31T23:59:59.000Z",
        }

        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await data_connectors.create(**requested_config)
        mock_proximl._query.assert_called_once_with(
            "/provider/provider-id-1/region/region-id-1/data_connector",
            "POST",
            None,
            expected_payload,
        )
        assert response.id == "connector-id-1"


class DataConnectorTests:
    def test_data_connector_properties(self, data_connector):
        assert isinstance(data_connector.id, str)
        assert isinstance(data_connector.provider_uuid, str)
        assert isinstance(data_connector.region_uuid, str)
        assert isinstance(data_connector.type, str)
        assert isinstance(data_connector.name, str)

    def test_data_connector_str(self, data_connector):
        string = str(data_connector)
        regex = r"^{.*\"connector_id\": \"" + data_connector.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_data_connector_repr(self, data_connector):
        string = repr(data_connector)
        regex = (
            r"^DataConnector\( proximl , \*\*{.*'connector_id': '"
            + data_connector.id
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_data_connector_bool(self, data_connector, mock_proximl):
        empty_data_connector = specimen.DataConnector(mock_proximl)
        assert bool(data_connector)
        assert not bool(empty_data_connector)

    @mark.asyncio
    async def test_data_connector_remove(self, data_connector, mock_proximl):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await data_connector.remove()
        mock_proximl._query.assert_called_once_with(
            "/provider/1/region/a/data_connector/x", "DELETE"
        )

    @mark.asyncio
    async def test_data_connector_refresh(self, data_connector, mock_proximl):
        api_response = {
            "provider_uuid": "provider-id-1",
            "region_uuid": "region-id-1",
            "connector_id": "connector-id-1",
            "name": "On-Prem DataConnector",
            "type": "custom",
            "cidr": "192.168.0.50/32",
            "port": "443",
            "protocol": "tcp",
            "createdAt": "2020-12-31T23:59:59.000Z",
        }
        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await data_connector.refresh()
        mock_proximl._query.assert_called_once_with(
            f"/provider/1/region/a/data_connector/x", "GET"
        )
        assert data_connector.id == "connector-id-1"
        assert response.id == "connector-id-1"
